@@ -7,7 +7,7 @@ use Config;
 use Text::ParseWords;
 
 use vars qw($VERSION);
-$VERSION = '0.00_01';
+$VERSION = '0.12';
 
 sub new {
   my $class = shift;
@@ -43,6 +43,28 @@ sub object_file {
   return "$file_base$self->{config}{obj_ext}";
 }
 
+sub arg_include_dirs {
+  my $self = shift;
+  return map {"-I$_"} @_;
+}
+
+sub arg_nolink { '-c' }
+
+sub arg_object_file {
+  my ($self, $file) = @_;
+  return ('-o', $file);
+}
+
+sub arg_share_object_file {
+  my ($self, $file) = @_;
+  return ($self->split_like_shell($self->{config}{lddlflags}), '-o', $file);
+}
+
+sub arg_exec_file {
+  my ($self, $file) = @_;
+  return ('-o', $file);
+}
+
 sub compile {
   my ($self, %args) = @_;
   die "Missing 'source' argument to compile()" unless defined $args{source};
@@ -51,18 +73,23 @@ sub compile {
 
   $args{object_file} ||= $self->object_file($args{source});
   
-  my @include_dirs = map {"-I$_"} (@{$args{include_dirs} || []},
-				   File::Spec->catdir($cf->{installarchlib}, 'CORE'));
+  my @include_dirs = $self->arg_include_dirs
+    (@{$args{include_dirs} || []},
+     $self->perl_inc());
   
   my @extra_compiler_flags = $self->split_like_shell($args{extra_compiler_flags});
   my @cccdlflags = $self->split_like_shell($cf->{cccdlflags});
   my @ccflags = $self->split_like_shell($cf->{ccflags});
   my @optimize = $self->split_like_shell($cf->{optimize});
-  my @flags = (@include_dirs, @cccdlflags, @extra_compiler_flags, '-c', @ccflags, @optimize);
+  my @flags = (@include_dirs, @cccdlflags, @extra_compiler_flags,
+	       $self->arg_nolink,
+	       @ccflags, @optimize,
+	       $self->arg_object_file($args{object_file}),
+	      );
   
   my @cc = $self->split_like_shell($cf->{cc});
   
-  $self->do_system(@cc, @flags, '-o', $args{object_file}, $args{source})
+  $self->do_system(@cc, @flags, $args{source})
     or die "error building $args{object_file} from '$args{source}'";
 
   return $args{object_file};
@@ -153,13 +180,13 @@ sub _do_link {
   my @temp_files;
   @temp_files =
     $self->prelink(%args,
-		   dl_name => $args{module_name}) if $self->need_prelink;
+		   dl_name => $args{module_name}) if $args{lddl} && $self->need_prelink;
   
   my @linker_flags = $self->split_like_shell($args{extra_linker_flags});
-  my @lddlflags = $args{lddl} ? $self->split_like_shell($cf->{lddlflags}) : ();
+  my @output = $args{lddl} ? $self->arg_share_object_file($out) : $self->arg_exec_file($out);
   my @shrp = $self->split_like_shell($cf->{shrpenv});
   my @ld = $self->split_like_shell($cf->{ld});
-  $self->do_system(@shrp, @ld, @lddlflags, '-o', $out, @$objects, @linker_flags)
+  $self->do_system(@shrp, @ld, @output, @$objects, @linker_flags)
     or die "error building $out from @$objects";
   
   return wantarray ? ($out, @temp_files) : $out;
@@ -168,7 +195,7 @@ sub _do_link {
 
 sub do_system {
   my ($self, @cmd) = @_;
-  print "@cmd\n";
+  print "@cmd\n" if !$self->{quiet};
   return !system(@cmd);
 }
 
@@ -183,4 +210,40 @@ sub split_like_shell {
   return Text::ParseWords::shellwords($string);
 }
 
+# if building perl, perl's main source directory
+sub perl_src {
+  # N.B. makemaker actually searches regardless of PERL_CORE, but
+  # only squawks at not finding it if PERL_CORE is set
+
+  return unless $ENV{PERL_CORE};
+
+  my $Updir  = File::Spec->updir;
+  my $dir = $Updir;
+
+  # Try up to 5 levels upwards
+  for (1..5) {
+    if (
+	-f File::Spec->catfile($dir,"config_h.SH")
+	&&
+	-f File::Spec->catfile($dir,"perl.h")
+	&&
+	-f File::Spec->catfile($dir,"lib","Exporter.pm")
+       ) {
+      return $dir;
+    }
+
+    $dir = File::Spec->catdir($dir, $Updir);
+  }
+  
+  warn "PERL_CORE is set but I can't find your perl source!\n";
+  return;
+}
+
+# directory of perl's include files
+sub perl_inc {
+  my $self = shift;
+
+  $self->perl_src() || File::Spec->catdir($self->{config}{archlibexp},"CORE");
+}
+
 1;
