@@ -8,7 +8,8 @@ use File::Spec;
 
 use ExtUtils::CBuilder::Base;
 
-use vars qw(@ISA);
+use vars qw($VERSION @ISA);
+$VERSION = '0.12';
 @ISA = qw(ExtUtils::CBuilder::Base);
 
 sub new {
@@ -32,18 +33,61 @@ sub _compiler_type {
 }
 
 sub split_like_shell {
-  # In Win32 command shells, the backslashes in the string "\bar\baz"
-  # should be preserved, because they often show up as parts of
-  # pathnames.  We double-up all these backslashes so shellwords() can
-  # still be used (making sure to avoid backslash-quote pairs, which
-  # should work the same as on Unix).
-
+  # As it turns out, Windows command-parsing is very different from
+  # Unix command-parsing.  Double-quotes mean different things,
+  # backslashes don't necessarily mean escapes, and so on.  So we
+  # can't use Text::ParseWords::shellwords() to break a command string
+  # into words.  The algorithm below was bashed out by Randy and Ken
+  # (mostly Randy), and there are a lot of regression tests, so we
+  # should feel free to adjust if desired.
+  
   (my $self, local $_) = @_;
-  if (defined() && length() && !ref()) {
-    s/\\(?!")/\\\\/g;
+  
+  return @$_ if defined() && UNIVERSAL::isa($_, 'ARRAY');
+  
+  my @argv;
+  return @argv unless defined() && length();
+  
+  my $arg = '';
+  my( $i, $quote_mode ) = ( 0, 0 );
+  
+  while ( $i < length() ) {
+    
+    my $ch      = substr( $_, $i  , 1 );
+    my $next_ch = substr( $_, $i+1, 1 );
+    
+    if ( $ch eq '\\' && $next_ch eq '"' ) {
+      $arg .= '"';
+      $i++;
+    } elsif ( $ch eq '\\' && $next_ch eq '\\' ) {
+      $arg .= '\\';
+      $i++;
+    } elsif ( $ch eq '"' && $next_ch eq '"' && $quote_mode ) {
+      $quote_mode = !$quote_mode;
+      $arg .= '"';
+      $i++;
+    } elsif ( $ch eq '"' && $next_ch eq '"' && !$quote_mode &&
+	      ( $i + 2 == length()  ||
+		substr( $_, $i + 2, 1 ) eq ' ' )
+	    ) { # for cases like: a"" => [ 'a' ]
+      push( @argv, $arg );
+      $arg = '';
+      $i += 2;
+    } elsif ( $ch eq '"' ) {
+      $quote_mode = !$quote_mode;
+    } elsif ( $ch eq ' ' && !$quote_mode ) {
+      push( @argv, $arg ) if $arg;
+      $arg = '';
+      ++$i while substr( $_, $i + 1, 1 ) eq ' ';
+    } else {
+      $arg .= $ch;
+    }
+    
+    $i++;
   }
-
-  return $self->SUPER::split_like_shell($_);
+  
+  push( @argv, $arg ) if defined( $arg ) && length( $arg );
+  return @argv;
 }
 
 sub compile {
@@ -70,9 +114,9 @@ sub compile {
                    ],
     optimize    => [ $self->split_like_shell($cf->{optimize})    ],
     defines     => [ '' ],
-    includes    => $args{include_dirs} || [],
+    includes    => [ @{$args{include_dirs} || []} ],
     perlinc     => [
-                     File::Spec->catdir($cf->{archlib}, 'CORE'),
+                     $self->perl_inc(),
                      $self->split_like_shell($cf->{incpath}),
                    ],
     use_scripts => 1, # XXX provide user option to change this???
@@ -108,6 +152,13 @@ sub link {
   $to ||= File::Spec->curdir();
 
   (my $file_base = $args{module_name}) =~ s/.*:://;
+  my $output = $args{lib_file} ||
+    File::Spec->catfile($to, "$file_base.$cf->{dlext}");
+
+  # if running in perl source tree, look for libs there, not installed
+  my $lddlflags = $cf->{lddlflags};
+  my $perl_src = $self->perl_src();
+  $lddlflags =~ s/\Q$cf->{archlibexp}\E[\\\/]CORE/$perl_src/ if $perl_src;
 
   my %spec = (
     srcdir        => $to,
@@ -115,12 +166,12 @@ sub link {
     startup       => [ ],
     objects       => \@objects,
     libs          => [ ],
-    output        => File::Spec->catfile($to, "$file_base.$cf->{dlext}"),
+    output        => $output,
     ld            => $cf->{ld},
     libperl       => $cf->{libperl},
     perllibs      => [ $self->split_like_shell($cf->{perllibs})  ],
     libpath       => [ $self->split_like_shell($cf->{libpth})    ],
-    lddlflags     => [ $self->split_like_shell($cf->{lddlflags}) ],
+    lddlflags     => [ $self->split_like_shell($lddlflags) ],
     other_ldflags => [ $self->split_like_shell($args{extra_linker_flags} || '') ],
     use_scripts   => 1, # XXX provide user option to change this???
   );
@@ -191,6 +242,19 @@ sub normalize_filespecs {
   }
 }
 
+# directory of perl's include files
+sub perl_inc {
+  my $self = shift;
+
+  my $perl_src = $self->perl_src();
+
+  if ($perl_src) {
+    File::Spec->catdir($perl_src, "lib", "CORE");
+  } else {
+    File::Spec->catdir($self->{config}{archlibexp},"CORE");
+  }
+}
+
 1;
 
 ########################################################################
@@ -247,7 +311,7 @@ sub write_compiler_script {
 
   $self->add_to_cleanup($script);
 
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
@@ -305,7 +369,7 @@ sub write_linker_script {
 
   $self->add_to_cleanup($script);
 
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
@@ -362,7 +426,7 @@ sub write_compiler_script {
 
   $self->add_to_cleanup($script);
 
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
@@ -422,7 +486,7 @@ sub write_linker_script {
 
   $self->add_to_cleanup($ld_script, $ld_libs);
 
-  print "Generating scripts '$ld_script' and '$ld_libs'.\n";
+  print "Generating scripts '$ld_script' and '$ld_libs'.\n" if !$self->{quiet};
 
   # Script 1: contains options & names of object files.
   open( LD_SCRIPT, ">$ld_script" )
@@ -467,8 +531,11 @@ sub format_compiler_cmd {
     $path = '-I' . $path;
   }
 
+  # split off any -arguments included in cc
+  my @cc = split / (?=-)/, $spec{cc};
+
   return [ grep {defined && length} (
-    $spec{cc}, '-c'         ,
+    @cc, '-c'               ,
     @{$spec{includes}}      ,
     @{$spec{cflags}}        ,
     @{$spec{optimize}}      ,
@@ -513,8 +580,11 @@ sub format_linker_cmd {
                '--output-exp' , $spec{explib}
   ];
 
+  # split off any -arguments included in ld
+  my @ld = split / (?=-)/, $spec{ld};
+
   push @cmds, [ grep {defined && length} (
-    $spec{ld}                 ,
+    @ld                       ,
     '-o', $spec{output}       ,
     "-Wl,--base-file,$spec{base_file}"   ,
     "-Wl,--image-base,$spec{image_base}" ,
@@ -536,7 +606,7 @@ sub format_linker_cmd {
   ];
 
   push @cmds, [ grep {defined && length} (
-    $spec{ld}                 ,
+    @ld                       ,
     '-o', $spec{output}       ,
     "-Wl,--image-base,$spec{image_base}" ,
     @{$spec{lddlflags}}       ,
@@ -561,7 +631,7 @@ sub write_linker_script {
 
   $self->add_to_cleanup($script);
 
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
